@@ -99,30 +99,69 @@ QuantizationPosition prepareQuantizationPosition(const std::vector<Mesh>& meshes
 	return result;
 }
 
-void prepareQuantizationTexture(cgltf_data* data, std::vector<QuantizationTexture>& result, const std::vector<Mesh>& meshes, const Settings& settings)
+static size_t follow(std::vector<size_t>& parents, size_t index)
 {
+	while (index != parents[index])
+	{
+		size_t parent = parents[index];
+
+		parents[index] = parents[parent];
+		index = parent;
+	}
+
+	return index;
+}
+
+void prepareQuantizationTexture(cgltf_data* data, std::vector<QuantizationTexture>& result, std::vector<size_t>& indices, const std::vector<Mesh>& meshes, const Settings& settings)
+{
+	// use union-find to associate each material with a canonical material
+	// this is necessary because any set of materials that are used on the same mesh must use the same quantization
+	std::vector<size_t> parents(result.size());
+
+	for (size_t i = 0; i < parents.size(); ++i)
+		parents[i] = i;
+
+	for (size_t i = 0; i < meshes.size(); ++i)
+	{
+		const Mesh& mesh = meshes[i];
+
+		if (!mesh.material && mesh.variants.empty())
+			continue;
+
+		size_t root = follow(parents, (mesh.material ? mesh.material : mesh.variants[0].material) - data->materials);
+
+		for (size_t j = 0; j < mesh.variants.size(); ++j)
+		{
+			size_t var = follow(parents, mesh.variants[j].material - data->materials);
+
+			parents[var] = root;
+		}
+
+		indices[i] = root;
+	}
+
+	// compute canonical material bounds based on meshes that use them
 	std::vector<Bounds> bounds(result.size());
 
 	for (size_t i = 0; i < meshes.size(); ++i)
 	{
 		const Mesh& mesh = meshes[i];
 
-		if (!mesh.material)
+		if (!mesh.material && mesh.variants.empty())
 			continue;
 
-		size_t mi = mesh.material - data->materials;
-		assert(mi < bounds.size());
-
-		updateAttributeBounds(mesh, cgltf_attribute_type_texcoord, bounds[mi]);
+		indices[i] = follow(parents, indices[i]);
+		updateAttributeBounds(mesh, cgltf_attribute_type_texcoord, bounds[indices[i]]);
 	}
 
+	// update all material data using canonical bounds
 	for (size_t i = 0; i < result.size(); ++i)
 	{
 		QuantizationTexture& qt = result[i];
 
 		qt.bits = settings.tex_bits;
 
-		const Bounds& b = bounds[i];
+		const Bounds& b = bounds[follow(parents, i)];
 
 		if (b.isValid())
 		{
